module.exports = Visualizer = (function() {
  var THREE = require('three');
  var TrackballControls = require('three-trackballcontrols');

  Visualizer.create = function() {
    return new Visualizer();
  };

  Visualizer.prototype.renderObject = function( printModel, callback ) {
    var vis = new Visualizer();
    vis.init();

    vis.addPlane( printModel.printBedDimensions );
    // Shifts visualized print bed plane if necessary
    //vis.changePlane( printModel.printBedDimensions );

    if (typeof printModel.vizModel.geometryArr == 'object' && printModel.vizModel.geometryArr[0] != undefined) {
      colorScheme = printModel.driveColour;

      var secondLastGeo = [0, 0, 0, false, 0];
      var lastGeo = [0, 0, 0, false, 0];
      var currentGeo = [];

      var geometry = new THREE.Geometry();
      var lineMat = new THREE.LineBasicMaterial({
        color: 0xffffff,
        vertexColors: THREE.VertexColors,
        linewidth: 4
      });

      /** printModel.vizModel.geometryArr is an array full of arrays, each of which contains positional data pulled
          from the loaded g-code file, as well as an isExtruding? flag, and a specified tool number.

          This iterates through the array, and checks for certain conditions to be true, like whether
          two positions are extruding, which indicates that a line should be drawn between them. Or
          if it's a draw-travel-draw sequence, which indicates that a line will be necessary for the
          draw.

          Will need a re-visit as more prints are run through it, but seems to work great right now.
       */
      var geoArray = printModel.vizModel.geometryArr;

      geoArray.forEach( function(points) {
        currentGeo = points;

        if (currentGeo[3] == true && lastGeo[3] == true) {
          geometry.vertices.push( new THREE.Vector3( lastGeo[0], lastGeo[1], lastGeo[2] ) );
          geometry.vertices.push( new THREE.Vector3( currentGeo[0], currentGeo[1], currentGeo[2] ) );
          geometry.colors.push( new THREE.Color ( colorScheme[ lastGeo[4] ] ) );
          geometry.colors.push( new THREE.Color ( colorScheme[ currentGeo[4] ] ) );
          secondLastGeo = lastGeo;
          lastGeo = currentGeo;
        } else if (currentGeo[3] == true && lastGeo[3] == false && secondLastGeo[3] == true) {
          geometry.vertices.push( new THREE.Vector3( lastGeo[0], lastGeo[1], lastGeo[2] ) );
          geometry.vertices.push( new THREE.Vector3( currentGeo[0], currentGeo[1], currentGeo[2] ) );
          geometry.colors.push( new THREE.Color ( colorScheme[ lastGeo[4] ] ) );
          geometry.colors.push( new THREE.Color ( colorScheme[ currentGeo[4] ] ) );
          secondLastGeo = lastGeo;
          lastGeo = currentGeo;
        } else {
          secondLastGeo = lastGeo;
          lastGeo = currentGeo;
        }
      } );

      // Camera + Plane specific calculations
      geometry.computeBoundingBox();

      line = new THREE.LineSegments( geometry, lineMat );

      vis.camera.position.x = ( (geometry.boundingBox.max.x - geometry.boundingBox.min.x) / 2 ) + geometry.boundingBox.min.x;
      vis.camera.position.y = ( (geometry.boundingBox.max.y - geometry.boundingBox.min.y) / 2 ) + geometry.boundingBox.min.y;

      vis.scene.add(line);
    }

    vis.render();
    vis.animate();
  }

  function Visualizer() {
    this.isOriginBottomLeft = false;

    this.init();
    this.addPlane();
    this.render();
    this.animate();
  };

  Visualizer.prototype.addPlane = function( printBedDimensions ) {
    if ( printBedDimensions !== undefined && ( printBedDimensions.x !== 0 || printBedDimensions.y !== 0 ) ) {
      var plane = new THREE.Mesh(
        new THREE.PlaneGeometry( printBedDimensions.x, printBedDimensions.y, 10, 10 ),
        new THREE.MeshBasicMaterial( {
            color: 0xb1adad,
            wireframe: true
        } )
      );
    } else {
      var plane = new THREE.Mesh(
        new THREE.PlaneGeometry( 250, 140, 10, 10 ),
        new THREE.MeshBasicMaterial( {
            color: 0xb1adad,
            wireframe: true
        } )
      );
    }

    plane.name = "printBed";

    if ( this.isOriginBottomLeft == true ) {
      plane.translateX( 125 );
      plane.translateY( 70 );
    };

    this.scene.add( plane );
  };

  Visualizer.prototype.changePlane = function(printBedDimensions) {
    var _this = this;
    var plane = _this.scene.getObjectByName("printBed");
    if ( printBedDimensions.originMiddle == false ) {
      plane.translateX( printBedDimensions.x / 2 );
      plane.translateY( printBedDimensions.y / 2 );
    };

    console.log( _this );
    plane.updateMatrixWorld( true );
    //_this.scene.updateMatrixWorld( true );
  };

  Visualizer.prototype.init = function() {
    this.renderBox = document.querySelector("#vizBox");
    this.camera = new THREE.PerspectiveCamera( 75, window.innerWidth / window.innerHeight, 1, 1000 );
    this.controls = new TrackballControls( this.camera, this.renderBox );
    this.scene = new THREE.Scene();

    this.camera.position.z = 300;

    this.controls.target.set( 0, 0, 0 );
    this.controls.rotateSpeed = 2.0;
    this.controls.zoomSpeed = 2;
    this.controls.panSpeed = 2.0;

    this.controls.noZoom = false;
    this.controls.noPan = false;

    this.controls.staticMoving = true;
    this.controls.dynamicDampingFactor = 0.3;

    this.controls.keys = [ 65, 83, 68 ];

    this.controls.addEventListener( 'change', this.render.bind(this) );

    this.scene.fog = new THREE.FogExp2( 0xFFFFFF, 0.002 );

    var light = new THREE.DirectionalLight( 0xffffff );
    light.position.set( 1, 1, 1 );
    light.castShadow = true;
    this.scene.add( light );

    var light2 = new THREE.DirectionalLight( 0x002288 );
    light2.position.set( -1, -1, -1 );
    light2.castShadow = true;
    this.scene.add( light2 );

    var light3 = new THREE.AmbientLight( 0x222222 );
    this.scene.add( light3 );

    this.scene.rotation.x = -45 * Math.PI / 180;
    this.scene.rotation.z = -20 * Math.PI / 180;

    this.renderer = new THREE.WebGLRenderer();
    this.renderer.setClearColor( this.scene.fog.color );
    this.renderer.setPixelRatio( window.devicePixelRatio ? window.devicePixelRatio : 1 );
    this.renderer.setSize( window.innerWidth, window.innerHeight );
    this.renderer.setViewport( 0, 0, window.innerWidth, window.innerHeight );
    this.renderer.shadowMap.enabled = true;
    this.renderBox.innerHTML = "";
    this.renderBox.appendChild( this.renderer.domElement );
  };

  Visualizer.prototype.animate = function() {
    requestAnimationFrame( this.animate.bind(this) );
    this.controls.update();
  };

  Visualizer.prototype.render = function () {
    this.renderer.render( this.scene, this.camera );
  };

  return Visualizer;

})();
