var Print = require('./print');

var common = require('../common');
var temp = require('temp');
var fs = require('fs');
var es = require('event-stream');

/**
 * Taking care of all of the pre-parse preparations. Will involve parsing through the
 * g-code once, figuring out type of extrusion, printable heights, and the bounding
 * box of a print
 *
 * REPLACES:
 * findPrintedLayerHeights()
 * testRelative()
 * calculateFirstLayerBoundingBox()
 *
 * @return N/A
 */
Print.prototype.preFlight = function(callback) {
  var _this = this;
  var lastLayer = { x: 0, y: 0, z: 0, e: 0 };
  var currentLayer = { x: 0, y: 0, z: 0, e: 0 };
  var lineCount = 1; // LOC starts at line 1
  var regCount = 0;
  var allCount = 0;
  var postToolLineCount = 0;

  var cmdList = ["G0", "G1", "G92", "G91", "G90", "TOOLCHANGE"];
  var HeightTracker = require('../procedures/height-tracker');

  var readableStream = fs.createReadStream(_this.inputFile)
    .setEncoding('utf-8')
    .pipe(es.split())
    .pipe(es.mapSync(function(line){

      // Pause reading stream for processing
      readableStream.pause();

      (function(){

        // Processes each line, and streams to new gcode line
        var args = common.parseLine(line);
        if ( args && args.cmd != undefined && ~cmdList.indexOf(args.cmd) ) {
          HeightTracker[ args.cmd ](_this, args);
        }
        if ( args && args.cmd !== undefined && args.e !== undefined ) {
          if ( args.e > lastLayer.e ) {
            regCount += 1;
            lastLayer.e = args.e;
          }

          if ( args.cmd == 'G92' ) {
            lastLayer.e = args.e;
          }
          allCount += 1;
        }

        lineCount += 1;

        readableStream.resume();

      })();
    })
    .on('error', function(){
      console.log('Error while reading file.');
    })
    .on('end', function(){
      // Assumes that if 75% of all tested lines have greater extrusion values
      // than previous line, then it is a regular extrusion file, otherwise it is
      // relative extrusion
      if ( (regCount / allCount) > 0.75 ) {
        _this.relativeExtrude = false;
      } else {
        _this.relativeExtrude = true;
      }

      var layersCounted = Object.keys(_this.layerColourCount);

      layersCounted.forEach(function(key) {
        if (_this.layerColourCount[key][1] > 30) {
          _this.printLayers.push( parseFloat(key) );
        }
      });

      _this.printLayers.forEach(function(height) {
        if (_this.layerColourCount[height][0] > _this.maxColourChange) _this.maxColourChange = _this.layerColourCount[height][0];
      });

      //console.log("DONE! Regular: " + regCount + ", Total: " + allCount);
      //console.log("Relative: " + _this.relativeExtrude);
      //console.log(_this.printLayers);
      //console.log("Max Colour Changes: " + _this.maxColourChange);
      //console.log("Lines: " + lineCount);

      typeof callback === 'function' && callback();
    })
  );
};

Print.prototype.loadPrinterDefaults = function() {
  var _this = this;
  var printers = common.populateMachineList();

  if (printers.length !== 0) {
    _this.printBedDimensions.x = parseInt(printers[0][0].printBedX);
    _this.printBedDimensions.y = parseInt(printers[0][0].printBedY);
    _this.printBedDimensions.originMiddle = parseInt(printers[0][0].printBedOriginMiddle);
    _this.towerVolume = parseInt(printers[0][0].towerVolume);
    _this.towerFastRate = parseInt(printers[0][0].towerFastRate);
    _this.towerSlowRate = parseInt(printers[0][0].towerSlowRate);
  }
};

Print.prototype.parseGcode = function(callback) {
  var _this = this;
  var cmdList = ["G0", "G1", "G92", "G91", "G90", "TOOLCHANGE"];

  if (_this.relativeExtrude == true) {
    var Procedures = require('../procedures/relative');
  } else {
    var Procedures = require('../procedures/regular');
  }

  // Creates temporary write stream where the file will be loaded into so iterations
  // and operations can be performed on it
  var tempStream = temp.createWriteStream({suffix: '.gcode'});
  var readableStream = fs.createReadStream(_this.inputFile)
    .setEncoding('utf-8')
    .pipe(es.split())
    .pipe(es.mapSync(function(line){

      // Pause reading stream for processing
      readableStream.pause();

      (function(){

        tempStream.write(line + '\n');

        // Processes each line, and streams to new gcode line
        var args = common.parseLine(line);
        if ( args && args.cmd != undefined && ~cmdList.indexOf(args.cmd) ) {
          Procedures[ args.cmd ](_this, args);
        }

        // Resume reading stream
        readableStream.resume();

      })();
    })
    .on('error', function(){
      console.log('Error while reading file.');
    })
    .on('end', function(){
      // Ending splice length addition, doesn't get triggered because g-code has no ending
      if (_this.writeArray.length > 0) {
        _this.writeArray.push( [_this.oldTool, _this.Etotal] );
        _this.headerArray.spliceCount += 1;
        _this.headerArray.drivesUsed[_this.oldTool] = 1;
        _this.writeArray[_this.writeArray.length-1][1] += 850; // Ending splice buffer
      }

      // End stream and set tempFile property in model to the created path for later usage
      tempStream.end();
      _this.tempFile = tempStream.path;

      typeof callback === 'function' && callback();
    })
  );
};

Print.prototype.makeGcode = function(callback) {
  var _this = this;
  var gcodeStream = fs.createWriteStream(_this.outputFile + ".gcode");

  var readableStream = fs.createReadStream(_this.tempFile)
    .setEncoding('utf-8')
    .pipe(es.split())
    .pipe(es.mapSync(function(line){

      // Pause reading stream for processing
      readableStream.pause();

      (function(){
        if ( !line.match(/(T1|T2|T3)/) ) {
          gcodeStream.write(line + '\n');
        }

        // Resume reading stream
        readableStream.resume();

      })();
    })
    .on('error', function(){
      console.log('Error while reading file.');
    })
    .on('end', function(){
      gcodeStream.end();
      typeof callback === 'function' && callback();
    })
  );
};

Print.prototype.makeSEEMfile = function(callback) {
  var _this = this;

  if (_this.outputFile === undefined) return;

  var drivesUsedHex = _this.headerArray.drivesUsed.join(''); //.toString(16);

  // Creates process required to convert pulsesPerMM float to hex
  var ieee754 = require('ieee754');
  var ppm = _this.headerArray.pulsesPerMM;
  var pulsesPerMMHex = "";
  var buf = new Buffer(4);
  ieee754.write(buf, ppm, 0, false, 23, 4);

  for (var i = 0; i < buf.length; i++) {
    pulsesPerMMHex += ("00" + buf[i].toString(16)).substr(-2);
  }

  // Creates process required to convert stepsPerCount float to hex
  var ieee754 = require('ieee754');
  var spc = _this.headerArray.stepsPerCount;
  var stepsPerCountHex = "";
  var buf = new Buffer(4);
  ieee754.write(buf, spc, 0, false, 23, 4);

  for (var i = 0; i < buf.length; i++) {
    stepsPerCountHex += ("00" + buf[i].toString(16)).substr(-2);
  }

  var loadingOffsetHex = ("0000" + _this.headerArray.loadingOffset.toString(16)).substr(-4);
  var spliceDistHex = ("0000" + _this.headerArray.spliceDist.toString(16)).substr(-4);
  var cutterOptoDistHex = ("0000" + _this.headerArray.cutterOptoDist.toString(16)).substr(-4);
  var spliceCountHex = ("0000" + _this.headerArray.spliceCount.toString(16)).substr(-4);
  var pingCountHex = ("0000" + _this.headerArray.pingCount.toString(16)).substr(-4);

  // Writes header to file, using headerArray values
  fs.writeFileSync(_this.outputFile, "MSF1.1\r\n");
  fs.appendFileSync(_this.outputFile, "cu:" + drivesUsedHex + "\r\n");
  fs.appendFileSync(_this.outputFile, "ppm:" + pulsesPerMMHex + "\r\n");
  fs.appendFileSync(_this.outputFile, "lo:" + loadingOffsetHex + "\r\n");

  fs.appendFileSync(_this.outputFile, "spc:" + stepsPerCountHex + "\r\n");
  fs.appendFileSync(_this.outputFile, "sd:" + spliceDistHex + "\r\n");
  fs.appendFileSync(_this.outputFile, "od:" + cutterOptoDistHex + "\r\n");

  fs.appendFileSync(_this.outputFile, "ns:" + spliceCountHex + "\r\n");
  fs.appendFileSync(_this.outputFile, "np:" + pingCountHex + "\r\n");

  // Writes filament lengths to file, using writeArray values
  _this.writeArray.forEach(function(line) {
    var buf = new Buffer(4);
    var spliceLengthHex = "";
    ieee754.write(buf, line[1], 0, false, 23, 4);
    for (var i = 0; i < buf.length; i++) {
      spliceLengthHex += ("00" + buf[i].toString(16)).substr(-2);
    }
    writeString = "(" + ("00" + (line[0]).toString(16)).substr(-2) + "," + spliceLengthHex + ")";
    fs.appendFileSync(_this.outputFile, writeString.toString() + "\r\n")
  });

  // Write ping lengths to file, using writePingArray values
  _this.writePingArray.forEach(function(line) {
    var buf = new Buffer(4);
    var pingLengthHex = "";
    ieee754.write(buf, line[1], 0, false, 23, 4);
    for (var i = 0; i < buf.length; i++) {
      pingLengthHex += ("00" + buf[i].toString(16)).substr(-2);
    }
    writeString = "(" + ("00" + (line[0]).toString(16)).substr(-2) + "," + pingLengthHex + ")";
    fs.appendFileSync(_this.outputFile, writeString.toString() + "\r\n")
  });

  // DEBUG for verifying piece lengths
  _this.writeArray.forEach(function(line) {
    writeString = "(" + line[0] + "," + line[1] + ")";
    fs.appendFileSync(_this.outputFile + ".plain", writeString.toString() + "\r\n")
  });

  _this.writePingArray.forEach(function(line) {
    writeString = "(" + line[0] + "," + line[1] + ")";
    fs.appendFileSync(_this.outputFile + ".plain", writeString.toString() + "\r\n")
  });

  typeof callback === 'function' && callback();
};
