var Print = require('./print');

var common = require('../common');
var temp = require('temp');
var fs = require('fs');
var es = require('event-stream');

Print.prototype.calculateFirstLayerBoundingBox = function(callback) {
  var _this = this;
  var currentZ = 0;
  var lineCount = 0;
  var maxX = maxY = 0;
  var minX = minY = 1000;

  var readableStream = fs.createReadStream(_this.inputFile)
    .setEncoding('utf-8')
    .pipe(es.split())
    .pipe(es.mapSync(function(line){

      // Pause reading stream for processing
      readableStream.pause();

      (function(){

        // Processes each line, and streams to new gcode line
        var args = common.parseLine(line);

        if ( ~line.indexOf('M104 S0') && lineCount > 1500 ) { readableStream.destroy(); }

        if ( args && args.z ) {
          currentZ = parseFloat(args.z);
        }

        if ( args && args.cmd && args.x && args.y && ~_this.printLayers.indexOf(currentZ) ) {
          if ( args.x < minX ) { minX = args.x; }
          if ( args.x > maxX ) { maxX = args.x; }
          if ( args.y < minY ) { minY = args.y; }
          if ( args.y > maxY ) { maxY = args.y; }
        }

        lineCount += 1;

        readableStream.resume();

      })();
    })
    .on('error', function(){
      console.log('Error while reading file.');
    })
    .on('close', function(){
      _this.boundingBox.x = minX;
      _this.boundingBox.y = minY + ((maxY - minY)/2);
      typeof callback === 'function' && callback();
    })
  );
};

/**
 * Calculation of a tower layer being generated for a tool change or layer
 * height change
 * @param  { array } towerStringArray contains the strings for the generated
 *                   tower layer
 * @return { float } Returns float with length of tower layer.
 */
Print.prototype.calculateTowerLayerLength = function(towerStringArray) {
  var _this = this;

  var towerLength = 0;

  towerStringArray.forEach(function(line) {
    var args = common.parseLine(line);
    if ( args.e ) {
      if (_this.relativeExtrude == true) {
        towerLength += parseFloat(args.e);
      } else {
        if (args.cmd == 'G1') { towerLength = parseFloat(args.e); }
      }
    }
  });

  return towerLength;
};

Print.prototype.calculateFirstPieceOffset = function(towerStringArray) {
  var _this = this;

  var towerLength = 0;

  towerStringArray.forEach(function(line) {
    var args = common.parseLine(line);
    if ( args.e ) {
      if (_this.relativeExtrude == true) {
        towerLength += parseFloat(args.e);
      } else {
        if (args.cmd == 'G1') { towerLength = parseFloat(args.e); }
      }
    }
  });

  return (towerLength * 0.45);
};

Print.prototype.towerToggle = function(callback) {
  // Creates a tower using the Print model, which is passed into the
  // function using `this`
  var _this = this;
  var tower = require('../modules/tower').create(_this);
  _this.addTowersToGCode( function() {
    _this.visualizePrint( function() {
      typeof callback === 'function' && callback();
    });
  } );
};

Print.prototype.addTowersToGCode = function(callback) {
  var _this = this;
  temp.track();
  var gcodeStream = temp.createWriteStream({suffix: '.gcode'});

  var firstTool = _this.firstTool;
  var tower = require('../modules/tower').create(_this);

  var towerArray = [{
    x: _this.boundingBox.x-7,
    y: _this.boundingBox.y+10,
    flippedDraw: false,
    firstLayer: true,
    densityPerLayer: 1,
    isPinging: false
  },
  {
    x: _this.boundingBox.x-7,
    y: _this.boundingBox.y-40,
    flippedDraw: false,
    firstLayer: true,
    densityPerLayer: 1,
    isPinging: false
  },
  {
    x: _this.boundingBox.x-57,
    y: _this.boundingBox.y+10,
    flippedDraw: false,
    firstLayer: true,
    densityPerLayer: 1,
    isPinging: false
  },
  {
    x: _this.boundingBox.x-57,
    y: _this.boundingBox.y-40,
    flippedDraw: false,
    firstLayer: true,
    densityPerLayer: 1,
    isPinging: false
  }];

  var toolCount = 1;
  var lastZ = 0;
  var currentZ = 0;
  var currentLength = 0;
  var previousPingLength = 0;
  var nextPingValue = 0;
  var lastLayer = { x: 0, y: 0, z: 0, e: 0 };

  toolCountArray = [];

  var cmdList = ["G0", "G1", "G92", "G91", "G90", "TOOLCHANGE"];

  var readableStream = fs.createReadStream(_this.tempFile)
    .setEncoding('utf-8')
    .pipe(es.split())
    .pipe(es.mapSync(function(line){

      // Pause reading stream for processing
      readableStream.pause();

      (function(){

        // TODO: Revisit this logic - it seems a little too crazy
        var args = common.parseLine(line);

        if ( args && args.cmd != undefined && args.e != undefined ) {
          _this.currentLayer = args;
          if ( _this.relativeExtrude == true) {
            currentLength += args.e;
          } else {
            if (args.e <= 0 && lastLayer.e > args.e) {

            } else {
              currentLength += (args.e - lastLayer.e);
              lastLayer.e = args.e;
            }
          }

          if ( args.cmd == 'G92' && args.e != undefined ) {
            lastLayer.e = args.e;
          }
        }

        if ( args && args.z ) {
          lastZ = currentZ;
          currentZ = args.z;
        }

        // Checks for layer height change, and if a change is detected AND it isn't the first tool change,
        // we trigger this check.
        //
        // OBJECTIVE: if layer height occurs and not all of the towers had a layer printed, then it generates
        // a tower layer with lower density on every tower necessary
        if ( args && args.z && ~_this.printLayers.indexOf(args.z) ) {
          if ( toolCountArray[currentZ] == undefined ) { toolCountArray[currentZ] = 0; }

          for (var i = toolCountArray[lastZ]; i < _this.maxColourChange; i++) {
            // Ternary operator for selecting which density to use for a tower layer, only having it
            // be 100% infill if it's the first layer of the tower
            towerArray[toolCountArray[lastZ]].firstLayer == true ?
              towerArray[toolCountArray[lastZ]].densityPerLayer = 1 :
              towerArray[toolCountArray[lastZ]].densityPerLayer = 0.1;

            if ( previousPingLength * 1.15 > previousPingLength + 350 ) {
              nextPingValue = previousPingLength * 1.15;
            } else {
              nextPingValue = previousPingLength + 350;
            }

            if ( currentLength > nextPingValue ) {
              towerArray[toolCountArray[lastZ]].isPinging = true;
              previousPingLength = currentLength;
              _this.writePingArray.push( [100, currentLength] );
              _this.headerArray.pingCount += 1;
            }

            towerStringArray = tower.transitionLayer( towerArray[toolCountArray[lastZ]], _this.currentLayer.e, toolCount );
            if ( _this.writeHeightArray[toolCount-1] == undefined ) { _this.writeHeightArray[toolCount-1] = 0; }

            var towerLayerLength = _this.calculateTowerLayerLength(towerStringArray);
            //_this.writeTowerArray[toolCount] += towerLayerLength;
            _this.writeHeightArray[toolCount-1] += towerLayerLength;
            currentLength += towerLayerLength;

            if ( towerArray[toolCountArray[lastZ]].firstLayer == true ) {
              _this.pieceOffset = _this.calculateFirstPieceOffset(towerStringArray);
            }

            towerArray[toolCountArray[lastZ]].flippedDraw = !towerArray[toolCountArray[lastZ]].flippedDraw;
            towerArray[toolCountArray[lastZ]].firstLayer = false;
            towerStringArray.forEach(function(towerString) {
              gcodeStream.write(towerString);
            });
            toolCountArray[lastZ] += 1;
          }
          lastZ = currentZ;
        }

        // LOGIC: This entire section looks at a line, and takes action based on what it runs into
        // This section checks whether or not the line is a toolchange or a layer height change, and if it is,
        // checks whether or not it's the first tool change, then flips that bool so it moves to the next
        // conditional next time
        //
        // Next time around, as long as the tool is T0, and doesn't have the weird M commands, it makes a tower
        // if the tower flag is true
        //
        // If the line does have T0/T1/T2/T3, it checks for a tool change and creates a tower as necessary
        gcodeStream.write(line + '\n');

        if ( line.match(/(T0|T1|T2|T3)/) && !line.match(/(M6|M104|M109|M126|M127|M133|M134|M140)/) && firstTool == true && ~_this.printLayers.indexOf(currentZ) ) {
          firstTool = false;
        } else if ( line.match(/(T0|T1|T2|T3)/) && !line.match(/(M6|M104|M109|M126|M127|M133|M134|M140)/) && ~_this.printLayers.indexOf(currentZ) ) {
          if ( toolCountArray[currentZ] == undefined ) { toolCountArray[currentZ] = 0; }
          towerArray[toolCountArray[currentZ]].densityPerLayer = 1;

          if ( previousPingLength * 1.15 > previousPingLength + 350 ) {
            nextPingValue = previousPingLength * 1.15;
          } else {
            nextPingValue = previousPingLength + 350;
          }

          if ( currentLength > nextPingValue ) {
            towerArray[toolCountArray[currentZ]].isPinging = true;
            previousPingLength = currentLength;
            _this.writePingArray.push( [100, currentLength] );
            _this.headerArray.pingCount += 1;
          }

          towerStringArray = tower.transitionLayer( towerArray[toolCountArray[currentZ]], _this.currentLayer.e, toolCount );
          if ( _this.writeTowerArray[toolCount] == undefined ) { _this.writeTowerArray[toolCount] = 0; }

          var towerLayerLength = _this.calculateTowerLayerLength(towerStringArray);
          _this.writeTowerArray[toolCount] += towerLayerLength;
          currentLength += towerLayerLength;

          if ( towerArray[toolCountArray[currentZ]].firstLayer == true ) {
            _this.pieceOffset = _this.calculateFirstPieceOffset(towerStringArray);
          }

          towerArray[toolCountArray[currentZ]].flippedDraw = !towerArray[toolCountArray[currentZ]].flippedDraw;
          towerArray[toolCountArray[currentZ]].firstLayer = false;
          towerStringArray.forEach(function(towerString) {
            gcodeStream.write(towerString);
          });
          toolCountArray[currentZ] += 1;
          toolCount++;
        }

        // Resume reading stream
        readableStream.resume();

      })();
    })
    .on('error', function(){
      console.log('Error while reading file.');
    })
    .on('end', function(){
      gcodeStream.end();
      _this.tempFile = gcodeStream.path;

      var cumulativeTowerLayerLengths = 0;
      var cumulativeHeightTowerLayerLengths = 0;

      // Starts at index 1 in writeTowerArray (because we only want to change everything after
      // the first filament) and works its way through to modify the splice lengths inside of
      // writeArray

      for ( var index = 0; index < _this.writeArray.length; index++ ) {
        if ( index == _this.writeArray.length ) { break; }
        if ( _this.writeHeightArray[index] == undefined ) { cumulativeHeightTowerLayerLengths += 0; }
        if ( _this.writeHeightArray[index] ) {
          cumulativeHeightTowerLayerLengths += _this.writeHeightArray[index];
        }
        _this.writeArray[index][1] += cumulativeHeightTowerLayerLengths;
      };

      for ( var index = 1; index < _this.writeArray.length; index++ ) {
        if ( index == _this.writeTowerArray.length ) { break; }
        if ( _this.writeArray[index] == undefined ) { break; }
        cumulativeTowerLayerLengths += _this.writeTowerArray[index];
        _this.writeArray[index][1] += cumulativeTowerLayerLengths;
      };

      for ( var index = 0; index < _this.writeArray.length; index++ ) {
        _this.writeArray[index][1] += _this.pieceOffset;
      };

      typeof callback === 'function' && callback();
    })
  );
};